// Copyright 2017 Darius Neatu (neatudarius@gmail.com)

#include <allocator.h>

/* citeste si proceseaza cate o linie de la stdin */
/* returneaza 0 doar cand se ajunge la FINALIZE */
int readAction(Arena *arena) {
    char line[50];
    fgets(line, 30, stdin);

    const char *delim = " \n";
    int n, index, size, value, align, length;

    char *token = strtok(line, delim);
    if (!strcmp(token, "INITIALIZE")) {
        token = strtok(NULL, delim);
        n = atoi(token);
        INITIALIZE(arena, n);
    } else if (!strcmp(token, "FINALIZE")) {
        FINALIZE(arena);
        return 0;
    } else if (!strcmp(token, "DUMP")) {
        DUMP(arena);
    } else if (!strcmp(token, "ALLOC")) {
        token = strtok(NULL, delim);
        size = atoi(token);
        ALLOC(arena, size, 0);
    } else if (!strcmp(token, "FREE")) {
        token = strtok(NULL, delim);
        index = atoi(token);
        FREE(arena, index);
    } else if (!strcmp(token, "FILL")) {
        token = strtok(NULL, delim);
        index = atoi(token);
        token = strtok(NULL, delim);
        size = atoi(token);
        token = strtok(NULL, delim);
        value = atoi(token);
        FILL(arena, index, size, value);
    } else if (!strcmp(token, "SHOW")) {
        token = strtok(NULL, delim);
        if (!strcmp(token, "FREE")) {
            SHOW_FREE(arena);
        } else if (!strcmp(token, "USAGE")) {
            SHOW_USAGE(arena);
        } else if (!strcmp(token, "ALLOCATIONS")) {
            SHOW_ALLOCATIONS(arena);
        } else if (!strcmp(token, "MAP")) {
            token = strtok(NULL, delim);
            length = atoi(token);
            SHOW_MAP(arena, length);
        }
    } else if (!strcmp(token, "ALLOCALIGNED")) {
        token = strtok(NULL, delim);
        size = atoi(token);
        token = strtok(NULL, delim);
        align = atoi(token);
        ALLOCALIGNED(arena, size, 0, align);
    } else if (!strcmp(token, "REALLOC")) {
        token = strtok(NULL, delim);
        index = atoi(token);
        token = strtok(NULL, delim);
        size = atoi(token);
        REALLOC(arena, index, size);
    }
    return 1;
}

/* aloca si seteaza la 0 n octeti, memorand adresa blocului in arena */
void INITIALIZE(Arena *arena, int n) {
    arena->addr = (unsigned char *)calloc(n, sizeof(unsigned char));
    arena->size = n;
}

/* elibereaza blocul de memorie folosit */
void FINALIZE(Arena *arena) {
    free(arena->addr);
    arena->size = 0;
}

/* afiseaza intreaga harta a memoriei */
void DUMP(Arena *arena) {
    int i;
    for (i = 0; i < arena->size; i++) {
        if (i % 16 == 0) {
            printf("%08X\t", i);
        }

        if ( (i/8) % 2 ) {
            printf(" %02X", (arena->addr)[i]);
        } else {
            printf("%02X ", (arena->addr)[i]);
        }
        if ((i % 16 == 15) && (i < arena->size - 1)) {
            printf("\n");
        }
    }
    printf("\n%08X\n", arena->size);
}

/* seteaza int-ul de la n cu valoarea value */
void SET_INT(Arena *arena, int n, int value) {
    *(int *)(arena->addr + n) = value;
}

/* analog pentru octet */
void SET_BYTE(Arena *arena, int n, unsigned char value) {
    *(arena->addr + n) = value;
}

/* returneaza int-ul de la byte-ul n */
int READ_INT(Arena *arena, int n) {
    return *(int *)(arena->addr + n);
}

/* analog pentru octetul n */
unsigned char READ_BYTE(Arena *arena, int n) {
    return *(arena->addr + n);
}

int ALLOC(Arena *arena, int size, int start) {
    int block_size, next_block, clear_start;

    /* urmatorul bloc, indicat de start */
    next_block = READ_INT(arena, start);

    /* PRIMUL pe care il aloc */
    if (start == 0 && next_block == 0) {
        clear_start = 4;
        if (arena->size - clear_start >= size + 12) {
            SET_INT(arena, 0, clear_start);
            SET_INT(arena, clear_start, 0);
            SET_INT(arena, clear_start+4, 0);
            SET_INT(arena, clear_start+8, 12+size);
            printf("%d\n", clear_start+12);
            return clear_start + 12;
        } else {
            printf("0\n");
            return 0;
        }
    } else if (start == 0 && next_block) {/*  Incerc intre inceput si primul */
        clear_start = 4;
        if (next_block - clear_start >= size + 12) {
            SET_INT(arena, 0, clear_start);
            SET_INT(arena, next_block+4, clear_start);
            SET_INT(arena, clear_start, next_block);
            SET_INT(arena, clear_start+4, 0);
            SET_INT(arena, clear_start+8, 12+size);
            printf("%d\n", clear_start+12);
            return clear_start + 12;
        } else {
            return ALLOC(arena, size, next_block);
        }
    } else if (start && next_block == 0) {
        block_size = READ_INT(arena, start+8);
        clear_start = start + block_size;
        if (arena->size - clear_start >= size + 12) {
            SET_INT(arena, start, clear_start);
            SET_INT(arena, clear_start, next_block);
            SET_INT(arena, clear_start+4, start);
            SET_INT(arena, clear_start+8, 12+size);
            printf("%d\n", clear_start+12);
            return clear_start + 12;
        } else {
            printf("0\n");
            return 0;
        }
    } else if (start && next_block) {   /* la mijloc */
        block_size = READ_INT(arena, start+8);
        clear_start = start + block_size;
        if (next_block - clear_start >= size + 12) {
            SET_INT(arena, start, clear_start);
            SET_INT(arena, next_block+4, clear_start);
            SET_INT(arena, clear_start, next_block);
            SET_INT(arena, clear_start+4, start);
            SET_INT(arena, clear_start+8, 12+size);
            printf("%d\n", clear_start+12);
            return clear_start + 12;
        } else {
            return ALLOC(arena, size, next_block);
        }
    }

    return 0; /* nu ajunge aici - anti-warning */
}

void FREE(Arena *arena, int index) {
    int next, prev;
    next = READ_INT(arena, index-12);
    prev = READ_INT(arena, index-8);
    if (next) {                         /* daca nu era ultimul */
        SET_INT(arena, next+4, prev);     /* actualizarea blocului urmator */
    }
    SET_INT(arena, prev, next);     /* actualizarea blocului precedent */
}

/* echivalentul memset */
void FILL(Arena *arena, int index, int size, unsigned char value) {
    for (int i = 0; i < size; i++) {
        SET_BYTE(arena, index + i, value);
    }
}

int NUMBER_OF_BLOCKS(Arena *arena) {   /* numarul de blocuri din arena */
    int next = 0, number = 0;
    while ((next = READ_INT(arena, next))) {
        number++;
    }
    return number;
}

int NUMBER_OF_BYTES(Arena *arena) {   /* numarul de octeti rezervati */
    int next = 0, number = 4; /* cei 4 de la inceputul arenei */
    while ((next = READ_INT(arena, next))) {
        number += READ_INT(arena, next+8);
    }
    return number;
}

/* afiseaza numarul de zone si de octeti liberi din arena */
void SHOW_FREE(Arena *arena) {
    int free_areas, free_bytes;
    free_bytes = arena->size - NUMBER_OF_BYTES(arena);

    int next = READ_INT(arena, 0), next_next;
    if (next == 0) {   /* nu am blocuri */
        free_areas = arena->size > 4 ? 1 : 0;
    } else {
        free_areas = next > 4 ? 1 : 0;
        do {
            next_next = READ_INT(arena, next);
            if (next_next) {
                int cnt = next_next > next + READ_INT(arena, next+8) ? 1 : 0;
                free_areas += cnt;
            } else {    /* capat */
                int cnt = arena->size > next + READ_INT(arena, next+8) ? 1 : 0;
                free_areas += cnt;
            }
        } while ((next = next_next));
    }

    printf("%d blocks (%d bytes) free\n", free_areas, free_bytes);
}

void SHOW_USAGE(Arena *arena) {
    int used_bytes, number_of_blocks, free_areas;
    number_of_blocks = NUMBER_OF_BLOCKS(arena);
    used_bytes = NUMBER_OF_BYTES(arena) - 12*number_of_blocks - 4;
    int efficiency = used_bytes * 100 / NUMBER_OF_BYTES(arena);

    int next = READ_INT(arena, 0), next_next;
    if (next == 0) {  /* nu am blocuri */
        free_areas = arena->size > 4 ? 1 : 0;
    } else {
        free_areas = next > 4 ? 1 : 0;
        do {
            next_next = READ_INT(arena, next);
            if (next_next) {
                int cnt = next_next > next + READ_INT(arena, next+8) ? 1 : 0;
                free_areas += cnt;
            } else {    /* capat */
                int cnt = arena->size > next + READ_INT(arena, next+8) ? 1 : 0;
                free_areas += cnt;
            }
        } while ((next = next_next));
    }

    int fragmentation = number_of_blocks ?
        (free_areas - 1) * 100 / number_of_blocks : 0;

    printf("%d blocks (%d bytes) used\n", number_of_blocks, used_bytes);
    printf("%d%% efficiency\n", efficiency);
    printf("%d%% fragmentation\n", fragmentation);
}

/* afiseaza zonele libere/ocupate, in ordinea in care apar */
void SHOW_ALLOCATIONS(Arena *arena) {
    int size;

    printf("OCCUPIED 4 bytes\n");

    int next = READ_INT(arena, 0), next_next;
    if (next == 0) {      /* daca nu exista niciun bloc */
        if (arena->size > 4) {
            printf("FREE %d bytes\n", arena->size - 4);
        }
    } else {          /* daca exista cel putin bloc */
        if (next > 4) {
            printf("FREE %d bytes\n", next - 4);
        }
        do {
            printf("OCCUPIED %d bytes\n", READ_INT(arena, next+8));
            next_next = READ_INT(arena, next);
            if (next_next) {
                if ((size = (next_next - (next + READ_INT(arena, next+8))))) {
                    printf("FREE %d bytes\n", size);
                }
            } else {    /* capat */
                if ((size = (arena->size - (next + READ_INT(arena, next+8))))) {
                    printf("FREE %d bytes\n", size);
                }
            }
        } while ((next = next_next));
    }
}

/* aloca asemanator cu alloc, dar shifteaza
clear_start la o val. divizibila cu align */
int ALLOCALIGNED(Arena *arena, int size, int start, int align) {
    int block_size, next_block, clear_start, future_data_start;

    next_block = READ_INT(arena, start);

    if (start == 0 && next_block == 0) {
        clear_start = 4;
        future_data_start = clear_start + 12;
        if (future_data_start % align) {
            future_data_start -= future_data_start % align;
            future_data_start += align;
        }
        clear_start = future_data_start - 12;
        if (arena->size - clear_start >= size + 12) {
            SET_INT(arena, 0, clear_start);
            SET_INT(arena, clear_start, 0);
            SET_INT(arena, clear_start+4, 0);
            SET_INT(arena, clear_start+8, 12+size);
            printf("%d\n", clear_start+12);
            return clear_start + 12;
        } else {
            printf("0\n");
            return 0;
        }
    } else if (start == 0 && next_block) {
        clear_start = 4;
        future_data_start = clear_start + 12;
        if (future_data_start % align) {
            future_data_start -= future_data_start % align;
            future_data_start += align;
        }
        clear_start = future_data_start - 12;
        if (next_block - clear_start >= size + 12) {
            SET_INT(arena, 0, clear_start);
            SET_INT(arena, next_block+4, clear_start);
            SET_INT(arena, clear_start, next_block);
            SET_INT(arena, clear_start+4, 0);
            SET_INT(arena, clear_start+8, 12+size);
            printf("%d\n", clear_start+12);
            return clear_start + 12;
        } else {
            return ALLOCALIGNED(arena, size, next_block, align);
        }
    } else if (start && next_block == 0) {
        block_size = READ_INT(arena, start+8);
        clear_start = start + block_size;
        future_data_start = clear_start + 12;
        if (future_data_start % align) {
            future_data_start -= future_data_start % align;
            future_data_start += align;
        }
        clear_start = future_data_start - 12;
        if (arena->size - clear_start >= size + 12) {
            SET_INT(arena, start, clear_start);
            SET_INT(arena, clear_start, next_block);
            SET_INT(arena, clear_start+4, start);
            SET_INT(arena, clear_start+8, 12+size);
            printf("%d\n", clear_start+12);
            return clear_start + 12;
        } else {
            printf("0\n");
            return 0;
        }
    } else if (start && next_block) {   /* la mijloc */
        block_size = READ_INT(arena, start+8);
        clear_start = start + block_size;
        future_data_start = clear_start + 12;
        if (future_data_start % align) {
            future_data_start -= future_data_start % align;
            future_data_start += align;
        }
        clear_start = future_data_start - 12;
        if (next_block - clear_start >= size + 12) {
            SET_INT(arena, start, clear_start);
            SET_INT(arena, next_block+4, clear_start);
            SET_INT(arena, clear_start, next_block);
            SET_INT(arena, clear_start+4, start);
            SET_INT(arena, clear_start+8, 12+size);
            printf("%d\n", clear_start+12);
            return clear_start + 12;
        } else {
            return ALLOCALIGNED(arena, size, next_block, align);
        }
    }

    return 0; /* nu ajunge aici - anti-warning */
}

/* realoca blocul de la ^index^, cu o noua dimensiune ^size^;
 truncheaza la nevoie */
int REALLOC(Arena *arena, int index, int size) {
    int block_addr = index - 12;
    int data_addr = index;
    int data_size = READ_INT(arena, block_addr+8) - 12;

    FREE(arena, data_addr);

    int new_data_addr = ALLOC(arena, size, 0);
    int i;

    /* copiere octet cu octet */
    for (i=0; i < (data_size < size ? data_size : size); i++) {
        SET_BYTE(arena, new_data_addr + i, READ_BYTE(arena, data_addr + i));
    }

    return new_data_addr;
}

/* afiseaza o harta a memoriei, scalata la lungimea length */
void SHOW_MAP(Arena *arena, int length) {
    double bytes_per_char = (double)arena->size / (double)length;
    double bytes = 0;

    int block, block_end, printed = 0;

    while (bytes < 4) { /* zona de gestiune de la inceputul arenei */
        printf("*");
        bytes += bytes_per_char;
        printed++;
        if ((printed) % 80 == 0) {
            printf("\n");
        }
    }

    block = READ_INT(arena, 0);
    do {
        while (bytes + bytes_per_char <= block) {
            printf(".");
            bytes += bytes_per_char;
            printed++;
            if ((printed) % 80 == 0) {
                printf("\n");
            }
        }
        block_end = block + READ_INT(arena, block+8);
        while (bytes < block_end) {   /* de la block la block_end */
            printf("*");
            bytes += bytes_per_char;
            printed++;
            if ((printed) % 80 == 0) {
                printf("\n");
            }
        }
        block = READ_INT(arena, block);
    } while (block);
    while (printed < length) {
        printf(".");
        printed++;
        if (printed % 80 == 0) {
            printf("\n");
        }
    }
    if (printed % 80) {
        printf("\n");
    }
}
